# FACTANAL.S - Extensions to S-Plus's factor analysis facilities.

# Copyright (c) 1996 by Radford M. Neal 
#
# Permission is granted for anyone to copy, use, or modify this program 
# for purposes of research or education, provided this copyright notice 
# is retained, and note is made of any changes that have been made. 
#
# This program is distributed without any warranty, express or implied.
# As this program was written for research purposes only, it has not been
# tested to the degree that would be advisable in any important application.
# All use of this program is entirely at the user's own risk.


# These functions do various things concerning factor analysis models,
# as represented by an S-Plus "factanal" object.


# FIND "RECOGNITION" MODEL FOR A FACTOR ANALYSIS MODEL.  The recognition
# model is derived on the assumption that the mean of the data is zero.

factanal.rec <- function (m)
{
  x <- solve(diag(m$uniquenesses) + m$loadings %*% t(m$loadings))
  list(
    weights = x %*% m$loadings,
    covariance =
      diag(rep(1,m$factors)) - t(m$loadings) %*% x %*% m$loadings
  )
}


# FIT A FACTOR ANALYSIS MODEL BY EM.

factanal.fit.em <- function (cmat, k, p, start)
{
  m <- list(method = "em", factors = k, correlation = cmat)

  if (is.null(start))
  { m$uniquenesses <- rep(1,p)
    m$loadings <- matrix(0.1*rnorm(p*k),p,k)
  }
  else
  { m$uniquenesses <- start$uniquenesses
    m$loadings <- start$loadings
  }

  for (i in 1:200)
  {
    x <- solve(diag(m$uniquenesses) + m$loadings %*% t(m$loadings))
    w <- x %*% m$loadings
    v <- diag(rep(1,k)) - t(m$loadings) %*% x %*% m$loadings

    m$loadings <- t( solve (t(w) %*% cmat %*% w + v, t(cmat %*% w)) )
    m$uniquenesses <- diag (cmat - cmat %*% w %*% t(m$loadings))

    print(m$uniquenesses)
  }

  class(m) <- "factanal"

  m
}


# Normalize a factor analysis model, by scaling variables to have
# variance one (according to the model), with the scaling factors
# recorded as an adjustment to "scale".

factanal.normalize <- function (m)
{
  v <- m$loadings %*% t(m$loadings) + m$uniquenesses
  m$scale <- m$scale * sqrt(diag(v))
  m$uniquenesses <- m$uniquenesses / diag(v)
  m$loadings <- diag(1/sqrt(diag(v))) %*% m$loadings 

  m
}


# De-normalize a factor analysis model, by applying the scale factor
# saved in "scale", then changing "scale" to one. 

factanal.denormalize <- function (m)
{
  m$uniquenesses <- m$uniquenesses * m$scale^2
  m$loadings <- diag(m$scale) %*% m$loadings 
  m$scale[] <- 1

  m
}


# Plot loadings for two models with two factors.  By default, each model's 
# loadings are plotted twice, for the two orderings of the factors.  This
# can be overrriden by setting w1 or w1 to 0 or 1 to select a single 
# plot for each model.  In any case, the signs are changed so that the sum 
# of the loadings for each factor is positive.  Since this can sometimes
# get things wrong, the s11..s22 parameters allow the signs to be flipped.

factanal.plot2 <- function(m1,m2,w1=NULL,w2=NULL,s11=1,s12=1,s21=1,s22=1)
{
  plot(c(-1,1),c(-1,1),type="n")
  points(0,0)

  if (s11*sum(m1$loadings[,1])<0) m1$loadings[,1] <- -m1$loadings[,1]
  if (s12*sum(m1$loadings[,2])<0) m1$loadings[,2] <- -m1$loadings[,2]

  if (is.null(w1) || w1==0)
  { lines(m1$loadings[,1],m1$loadings[,2],type="b",lty=1,pch="*")
  }
  if (is.null(w1) || w1==1)
  { lines(m1$loadings[,2],m1$loadings[,1],type="b",lty=1,pch="*")
  }

  if (s21*sum(m2$loadings[,1])<0) m2$loadings[,1] <- -m2$loadings[,1]
  if (s22*sum(m2$loadings[,2])<0) m2$loadings[,2] <- -m2$loadings[,2]

  if (is.null(w2) || w2==0)
  { lines(m2$loadings[,1],m2$loadings[,2],type="b",lty=2,pch="*")
  }
  if (is.null(w2) || w2==1)
  { lines(m2$loadings[,2],m2$loadings[,1],type="b",lty=2,pch="*")
  }
}


# Generate a factor analysis model at random.  It's not necessarily 
# normalized.

factanal.randmodel <- function (k, p)
{
  m <- list( center = rnorm(p), scale = rep(1,p), uniquenesses = rexp(p), 
             factors = k, loadings = matrix(rnorm(p*k),p,k))

  class(m) <- "factanal"

  m
}


# Generate data randomly from a factor analysis model.  The data is
# produced in original (not normalized) form.

factanal.gen <- function (m, n)
{ 
  p <- length(m$center)
  k <- m$factors
  r <- matrix(0,n,p)

  for (i in 1:n) 
  { r[i,] <- m$center + m$scale * (as.vector(m$loadings %*% rnorm(k)) 
                                     + rnorm(p) * sqrt(m$uniquenesses))
  }
  
  r
}
