# TEST.S - S functions used in testing the wake-sleep factor analysis method.

# Copyright (c) 1996 by Radford M. Neal 
#
# Permission is granted for anyone to copy, use, or modify this program 
# for purposes of research or education, provided this copyright notice 
# is retained, and note is made of any changes that have been made. 
#
# This program is distributed without any warranty, express or implied.
# As this program was written for research purposes only, it has not been
# tested to the degree that would be advisable in any important application.
# All use of this program is entirely at the user's own risk.


# Function to generate models and data for tests of factor analysis using the 
# wake-sleep algorithm.  The parameters give the number of factors and the
# number of variables to generate models and data for.  Ten such models are
# generated, saved in S-Plus variables of the form testk:p-i.model.  For
# each of the ten models, two datasets are generated, with 10 and 100 cases,
# and saved in the S-Plus variables testk:p-i.data10 and testk:p-i.data100,
# and in files of the same names.

gen <- function (k, p)
{

  for (i in 0:9)
  { 
    prefix <- paste("test",k,":",p,"-",i,sep="")
    cat("Generating",prefix,"\n")

    m <- factanal.normalize(factanal.randmodel(k,p))
    m$center[] <- 0
    m$scale[] <- 1

    assign(paste(prefix,".model",sep=""),m,where=1)

    for (n in c(10,500))
    { 
      d <- factanal.gen(m,n)

      assign(paste(prefix,".data",n,sep=""),d,where=1)
      write.table(d,sep=" ",file=paste(prefix,".data",n,sep=""))
    }
  }
}


# Do maximum likelihood factor analysis for generated datasets.

mle <- function (k, p)
{
  for (i in 0:9)
  { 
    prefix <- paste("test",k,":",p,"-",i,sep="")
    cat("Finding ML estimates for",prefix,"\n")

    for (n in c(10,500))
    { 
      m <- factanal(get(paste(prefix,".data",n,sep="")),factors=k,method="mle")

      assign(paste(prefix,".mle",n,sep=""),factanal.denormalize(m),where=1)
    }
  }
}


# Do maximum likelihodd factor analysis for one dataset using starting
# values taken from the end of a wake-sleep run.  The fitted model is
# stored with a "b" appended to the usual name.

mle1 <- function (k, p, i, n)
{
  prefix <- paste("test",k,":",p,"-",i,sep="")
  cat("Finding ML estimates for ",prefix,", size ",n,
      " starting with:\n",sep="")

  ws <- source(paste(prefix,".ws",n,sep=""))
  print(ws$uniquenesses)
  ws <- factanal.normalize(ws)
  print(ws$uniquenesses)

  m <- factanal(get(paste(prefix,".data",n,sep="")),factors=k,method="mle",
                start=matrix(factanal.normalize(ws)$uniquenesses,p,1))

  assign(paste(prefix,".mle",n,"b",sep=""),factanal.denormalize(m),where=1)
}


# Plot path of variances in wake-sleep learning, comparing against true values 
# and maximum likelihood estimates.  The variable "pgroup" should be set
# beforehand to the number of presentations in each group (ie, for each
# data point).

vplot <- function (k, p, i, n, len=300, pmle=T, ptrue=T, suffix="")
{
  prefix <- paste("test",k,":",p,"-",i,sep="")
  
  vtrc <- read.table(paste(prefix,".vtrc",n,sep=""))
  mod <- get(paste(prefix,".model",sep=""))
  mle <- get(paste(prefix,".mle",n,suffix,sep=""))

  plot(pgroup*c(0,len+15)/1000000,c(0,1.2),type="n",
   xlab="", ylab="Generative Variances")

  for (v in 1:p)
  { lines(pgroup*(0:len)/1000000,vtrc[1:(len+1),v])
    if (pmle)
    { lines(pgroup*c(0,len+1)/1000000,rep(mle$uniquenesses[v],2))
      text(pgroup*c(len+5)/1000000,mle$uniquenesses[v],v)
    }
    if (ptrue)
    { lines(pgroup*c(0,len+6)/1000000,rep(mod$uniquenesses[v],2),lty=2)
      text(pgroup*c(len+10)/1000000,mod$uniquenesses[v],v)
    }
  }
}


# Same as vplot, but with smoothing.

vplot.s <- function (k, p, i, n, len=300, pmle=T, ptrue=T)
{
  prefix <- paste("test",k,":",p,"-",i,sep="")
  
  vtrc <- read.table(paste(prefix,".vtrc",n,sep=""))
  mod <- get(paste(prefix,".model",sep=""))
  mle <- get(paste(prefix,".mle",n,sep=""))

  plot(pgroup*c(0,len+15)/1000000,c(0,1),type="n",
    xlab="",ylab="Generative Variances")

  for (v in 1:p)
  { lines(pgroup*(0:(len-4))/1000000,
     (vtrc[1:(len-3),v]+vtrc[2:(len-2),v]+vtrc[3:(len-1),v]+vtrc[4:len,v])/4)
    if (pmle)
    { lines(pgroup*c(0,len+1)/1000000,rep(mle$uniquenesses[v],2))
      text(pgroup*c(len+5)/1000000,mle$uniquenesses[v],v)
    }
    if (ptrue)
    { lines(pgroup*c(0,len+6)/1000000,rep(mod$uniquenesses[v],2),lty=2)
      text(pgroup*c(len+10)/1000000,mod$uniquenesses[v],v)
    }
  }
}


# Same as vplot, but for the Everitt data.

everitt.vplot <- function (s,len)
{
  vtrc <- read.table(paste("everitt.vtrc",s,sep=""))
  mle <- everitt.mle

  plot(pgroup*c(0,len+15)/1000000,c(0,1),type="n",
    xlab="",ylab="Generative Variances")

  for (v in 1:7)
  { lines(pgroup*(0:len)/1000000,vtrc[1:(len+1),v])
    lines(pgroup*c(0,len+1)/1000000,rep(mle$uniquenesses[v],2))
    text(pgroup*c(len+5)/1000000,mle$uniquenesses[v],v)
  }
}


# Plot path of generative weights in wake-sleep learning, comparing against 
# true values and maximum likelihood estimates.  This procedure really only
# works for one factor.  It tries to adjust the signs of all the weights 
# to match the true model; the rev parameters let one override its decisions.
# The variable "pgroup" should be set beforehand to the number of presentations
# in each group (ie, for each data point).

wplot <- function (k, p, i, n, len=300, pmle=T, ptrue=T, 
                   rev.mod=F, rev.mle=F, rev.ws=F)
{
  prefix <- paste("test",k,":",p,"-",i,sep="")
  
  wtrc <- read.table(paste(prefix,".wtrc",n,sep=""))
  mod <- get(paste(prefix,".model",sep=""))
  mle <- get(paste(prefix,".mle",n,sep=""))

  plot(pgroup*c(0,len+15)/1000000,c(-1.1,1.1),type="n",
    xlab="",ylab="Generative Weights")

  s.mod <- sign(sum(as.vector(mod$loadings)))
  if (rev.mod) s.mod <- - s.mod
  s.mle <- sign(sum(as.vector(mle$loadings)))
  if (rev.mle) s.mle <- - s.mle
  s.ws  <- sign(sum(wtrc[dim(wtrc)[1],]))
  if (rev.ws) s.ws <- - s.ws

  for (w in 1:(p*k))
  { lines(pgroup*(0:len)/1000000,s.mod*s.ws*wtrc[1:(len+1),w])
    if (pmle)
    { lines(pgroup*c(0,len+1)/1000000,
       rep(s.mod*s.mle*as.vector(mle$loadings)[w],2))
      text(pgroup*c(len+5)/1000000,
       s.mod*s.mle*as.vector(mle$loadings)[w],w)
    }
    if (ptrue)
    { lines(pgroup*c(0,len+6)/1000000,rep(as.vector(mod$loadings)[w],2),lty=2)
      text(pgroup*c(len+10)/1000000,as.vector(mod$loadings)[w],w) 
    }
  }
}


# Plot weights for 2-factor models.

plot2 <- function(k,p,i,n,w1=NULL,w2=NULL,s11=1,s12=1,s21=1,s22=1)
{
  prefix <- paste("test",k,":",p,"-",i,sep="")

  factanal.plot2(rotate(source(paste(prefix,".ws",n,sep=""))),
                 rotate(get(paste(prefix,".mle",n,sep=""))),
                 w1,w2,s11,s12,s21,s22)
}


# Functions for producing Postscript figures for the paper.

start.fig <- function(file)
{
  ps.options(horizontal = F, width = 6.1, height = 3.4, pointsize = 9)
  postscript(file, onefile = F, print.it = F)
}

end.fig <- function()
{
  dev.off()
}
