/* WS-FA.C - Factor analysis using the wake-sleep algorithm. */

/* Copyright (c) 1996 by Radford M. Neal 
 *
 * Permission is granted for anyone to copy, use, or modify this program 
 * for purposes of research or education, provided this copyright notice 
 * is retained, and note is made of any changes that have been made. 
 *
 * This program is distributed without any warranty, express or implied.
 * As this program was written for research purposes only, it has not been
 * tested to the degree that would be advisable in any important application.
 * All use of this program is entirely at the user's own risk.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>


#include "misc.h"
#include "rand.h"
#include "numin.h"


/* THE DATA. */

int N;			/* Number of cases */
int V;			/* Number of variables */

double *data;		/* Values of variables for all cases */


/* THE GENERATIVE MODEL. */

int F;			/* Number of latent factors */

double *gw;		/* Generative weights */
double *gb;		/* Generative biases */
double *gn;		/* Generative noise variances */
  

/* THE RECOGNITION MODEL. */

int corr;		/* Are correlation weighs present? */

double *rw;		/* Recognition weights */
double *rb;		/* Recognition biases */
double *rn;		/* Recognition noise variances */
double *rc;		/* Correlation weights */


/* LEARNING PARAMETERS. */

double gstep, rstep;	/* Stepsizes for weight updates */
double gdecay, rdecay;	/* Decay rates for variance estimates */

int rnoise, gnoise;	/* Really add noise (rather than simulate)? */

int bigrv;		/* Allow big recognition variances? */
int smallgv;		/* Allow small generative variances? */


/* VARIOUS OPTIONS, ETC. */

int echo;		/* Echo options? */

int norm;		/* Normalize data? */
double *mean, *scale;	/* Normalizations constants for variables */

int syst;		/* Visit training cases systematically (not randomly) */

int iterations;		/* Number of iterations to do */
int mod;		/* Modulus for dislaying trace */

int seed;		/* Random number seed */

char *data_spec;	/* Specification for data source */
char *model_file;	/* Name of model file */
char *var_trace;	/* Name of file to receive variance trace */
char *weight_trace;	/* Name of file to receive weight trace */

int release_time;	/* Time to release smallgv constraint, zero if never */


/* PROCEDURES. */

static void usage(void), wake_sleep(void), display_model(void), 
            save_model(char *);


/* MAIN PROGRAM. */

void main
( int argc,
  char **argv
)
{
  numin_source ns;
  char **a;
  int i, v;

  /* Set defaults. */

  mod = 0;
  seed = 1;

  corr = 1;
  rnoise = 1;
  gnoise = 1;
  bigrv = 1;
  smallgv = 1;
  norm = 1;
  syst = 1;
  echo = 1;

  model_file = 0;
  weight_trace = 0;
  var_trace = 0;

  /* Look for option flags. */

  for (a = argv+1; *a!=0 && (**a=='-' || **a=='+'); a++)
  {
    if (strcmp(*a+1,"corr")==0)        corr = **a=='+';
    else if (strcmp(*a+1,"rnoise")==0) rnoise = **a=='+';
    else if (strcmp(*a+1,"gnoise")==0) gnoise = **a=='+';
    else if (strcmp(*a+1,"bigrv")==0)  bigrv = **a=='+';
    else if (strcmp(*a+1,"norm")==0)   norm = **a=='+';
    else if (strcmp(*a+1,"syst")==0)   syst = **a=='+';
    else if (strcmp(*a+1,"echo")==0)   echo = **a=='+';
    else if (strcmp(*a+1,"smallgv")==0)
    { smallgv = **a=='+';
      if (!smallgv)
      { release_time = atoi(*(a+1));
        a += 1;
      }
    }
    else
    { usage();
    }
  }

  /* Look at numeric arguments. */

  if (*a==0 || **a<'0' || **a>'9' || (V = atoi(*a++))<=0) usage();

  if (*a==0 || **a<'0' || **a>'9' || (F = atoi(*a++))<=0) usage();

  if (*a==0 || **a<'0' || **a>'9' || (iterations = atoi(*a))<=0) usage();
  if (strchr(*a,'%')!=0 && (mod = atoi(strchr(*a,'%')+1))<=0) usage();
  a += 1;

  if (*a==0 || **a<'0' || **a>'9' || (gstep = rstep = atof(*a))<=0) usage();
  if (strchr(*a,'/')!=0 && (rstep = atof(strchr(*a,'/')+1))<=0) usage();
  a += 1;

  if (*a==0 || **a<'0' || **a>'9' || (gdecay = rdecay = atof(*a))<=0) usage();
  if (strchr(*a,'/')!=0 && (rdecay = atof(strchr(*a,'/')+1))<=0) usage();
  if (gdecay>=1 || rdecay>=1) usage();
  a += 1;

  if (*a!=0 && **a>='0' && **a<='9')
  { if ((seed = atoi(*a++))<=0) usage();
  }

  /* Look for data specification. */

  if (*a==0) usage();
  data_spec = *a++;

  /* Look for model and trace files */

  if (*a!=0) model_file   = *a++;
  if (*a!=0) var_trace    = *a++;
  if (*a!=0) weight_trace = *a++;
    
  /* Check that there are no more arguments. */

  if (*a!=0) usage();

  /* Echo arguments if desired. */

  if (echo)
  {
    printf("\nWAKE-SLEEP FACTOR ANALYSIS\n\n");

    printf(
    "Options:     %ccorr %crnoise %cgnoise %cnorm %csyst %cbigrv %csmallgv",
    "-+"[corr], "-+"[rnoise], "-+"[gnoise], "-+"[norm], "-+"[syst], 
    "-+"[bigrv], "-+"[smallgv]);

    if (!smallgv) printf(" %d",release_time);
    
    printf("\n\n");

    printf("Variables:   %d\n",V);
    printf("Factors:     %d\n\n",F);

    printf("Iterations:  %-7d\n",iterations);
    printf("Stepsizes:   %.10lf/%.10lf (g/r)\n",gstep,rstep);
    printf("Decay-rates: %.10lf/%.10lf (g/r)\n\n",gdecay,rdecay);

    printf("Random seed: %d\n\n",seed);

    printf("Data source: %s\n",data_spec);
    if (model_file) printf("Model file:  %s\n",model_file);
    
    printf("\n");
  }

  /* Read the data. */

  numin_spec(&ns,"data@1,0",1);
  numin_spec(&ns,data_spec,V);

  N = numin_start(&ns);

  if (echo)
  { printf("Number of cases to read: %d\n\n",N);
  }

  data = chk_alloc (N*V, sizeof *data);

  for (i = 0; i<N; i++)
  { numin_read(&ns,data+i*V);
  }

  numin_close(&ns);

  /* Allocate space. */

  gw = chk_alloc (V*F, sizeof *gw);
  gb = chk_alloc (V,   sizeof *gb);
  gn = chk_alloc (V,   sizeof *gn);

  rw = chk_alloc (V*F, sizeof *rw);
  rb = chk_alloc (F,   sizeof *rb);
  rn = chk_alloc (F,   sizeof *rn);

  if (corr) rc = chk_alloc (F*F, sizeof *rc);

  mean  = chk_alloc (V, sizeof *mean);
  scale = chk_alloc (V, sizeof *scale);

  /* Normalize data, if desired. */

  if (norm)
  {
    if (echo) 
    { printf("Normalizations for variables:\n\n");
      printf("        Mean        Scale\n\n");
    }

    for (v = 0; v<V; v++)
    { 
      mean[v] = 0;
      for (i = 0; i<N; i++)
      { mean[v] += data[i*V+v];
      }
      mean[v] /= N;
  
      scale[v] = 0;
      for (i = 0; i<N; i++)
      { scale[v] += (data[i*V+v]-mean[v]) * (data[i*V+v]-mean[v]);
      }
      scale[v] = sqrt(scale[v]/N);
  
      for (i = 0; i<N; i++)
      { data[i*V+v] = (data[i*V+v] - mean[v]) / scale[v];
      }

      if (echo) printf("%3d  %10.5lf  %10.5lf\n",v+1,mean[v],scale[v]);
    }

    if (echo) printf("\n");
  }

  /* Do wake-sleep learning. */

  wake_sleep();

  /* Save final model in S-Plus format. */

  if (model_file)
  { save_model(model_file);
  }

  exit(0);
}


/* WAKE-SLEEP LEARNING ALGORITHM. */

static void wake_sleep(void)
{
  FILE *vtrc, *wtrc;	/* Trace files */

  double *fv;		/* Values for latent factors (real or fantasy case) */
  double *vv;		/* Values for visible variables (fantasy case) */

  int i, t, v, f, f2;
  double d, s;

  /* Set things up. */

  rand_seed(seed);

  fv = chk_alloc (F, sizeof *fv);
  vv = chk_alloc (V, sizeof *vv);

  if (var_trace)
  { vtrc = fopen(var_trace,"w"); 
    if (vtrc==NULL) 
    { fprintf(stderr,"Can't create variance trace file (%s)\n",var_trace);
      exit(1);
    }
  }

  if (weight_trace)
  { wtrc = fopen(weight_trace,"w"); 
    if (wtrc==NULL) 
    { fprintf(stderr,"Can't create weight trace file (%s)\n",weight_trace);
      exit(1);
    }
  }

  /* Initialize generative model. */

  for (v = 0; v<V; v++)
  { for (f = 0; f<F; f++) 
    { gw[v*F+f] = rnoise || gnoise ? 0 : 0.1*rand_gaussian();
    }
    gb[v] = 0;
    gn[v] = 1;
  }

  /* Initialize recognition model. */

  for (f = 0; f<F; f++)
  { for (v = 0; v<V; v++) rw[v*F+f] = 0;
    if (corr) 
    { for (f2 = 0; f2<f; f2++) rc[f2*F+f] = 0;
    }
    rb[f] = 0;
    rn[f] = 1;
  }

  /* Do learning iterations. */

  t = 0;

  for (;;)
  {
    /* Display model if desired. */

    if (mod && t%mod==0)
    {
      printf("\nITERATION %d\n\n",t);
      if (model_file)  
      { save_model(model_file);
      }
      display_model();
      fflush(stdout);

      if (var_trace)
      { for (v = 0; v<V; v++) 
        { fprintf(vtrc," %.4le",gn[v]);
        }
        fprintf(vtrc,"\n");
        fflush(vtrc);
      }

      if (weight_trace)
      { for (f = 0; f<F; f++)
        { for (v = 0; v<V; v++)
          { fprintf(wtrc," %+.5lf",gw[v*F+f]);
          }
        } 
        fprintf(wtrc,"\n");
        fflush(wtrc);
      }
    }

    /* See if we're done. */

    if (t==iterations) break;
    t += 1;

    /* Apply the recognition model to a training case selected systematically
       or at random. */

    i = syst ? (t-1)%N : rand_int(N);

    for (f = 0; f<F; f++)
    { fv[f] = rb[f];
      for (v = 0; v<V; v++)
      { fv[f] += rw[v*F+f] * data[i*V+v];
      }
      if (corr)
      { for (f2 = 0; f2<f; f2++)
        { fv[f] += rc[f2*F+f] * fv[f2];
        }   
      }
      if (rnoise) fv[f] += sqrt(rn[f]) * rand_gaussian();
    }

    /* Do wake-phase learning for the generative model. */

    for (v = 0; v<V; v++)
    { d = gb[v];
      for (f = 0; f<F; f++)
      { d += gw[v*F+f] * fv[f];
      }
      s = 0;
      if (!rnoise)
      { for (f = 0; f<F; f++)
        { s += gw[v*F+f]*gw[v*F+f]*rn[f];
        }
      }
      d -= data[i*V+v];
      for (f = 0; f<F; f++)
      { gw[v*F+f] -= gstep * (fv[f] * d + (rnoise ? 0 : gw[v*F+f] * rn[f]));
      }
      if (!norm) gb[v] -= gstep * d;
      gn[v] = gdecay * gn[v] + (1-gdecay) * (d*d + s);
      if (!smallgv && (release_time==0 || t<=release_time) && gn[v]<0.01) 
      { gn[v] = 0.01;
      }
    }

    /* Generate a fantasy using the generative model. */

    for (f = 0; f<F; f++)
    { fv[f] = rand_gaussian();
    }

    for (v = 0; v<V; v++)
    { vv[v] = gb[v];
      for (f = 0; f<F; f++)
      { vv[v] += gw[v*F+f] * fv[f];
      }
      if (gnoise) vv[v] += sqrt(gn[v]) * rand_gaussian();
    }

    /* Do sleep-phase learning for the recognition model. */

    for (f = 0; f<F; f++)
    { d = rb[f];
      for (v = 0; v<V; v++)
      { d += rw[v*F+f] * vv[v];
      }
      s = 0;
      if (!gnoise)
      { for (v = 0; v<V; v++)
        { s += rw[v*F+f]*rw[v*F+f]*gn[v];
        }
      }
      if (corr)
      { for (f2 = 0; f2<f; f2++)
        { d += rc[f2*F+f] * fv[f2];
        }
      }
      d -= fv[f];
      for (v = 0; v<V; v++)
      { rw[v*F+f] -= rstep * (vv[v] * d + (gnoise ? 0 : rw[v*F+f] * gn[v]));
      }
      if (corr)
      { for (f2 = 0; f2<f; f2++)
        { rc[f2*F+f] -= rstep * fv[f2] * d;
        }
      }
      if (!norm) rb[f] -= rstep * d;
      rn[f] = rdecay * rn[f] + (1-rdecay) * (d*d + s);
      if (rn[f]>1 && !bigrv) rn[f] = 1;
    }
  }
}


/* DISPLAY CURRENT MODEL ON STANDARD OUTPUT. */

static void display_model(void)
{
  int v, f, f2;

  printf("Generative model:\n\n");
  if (!norm)
  { printf("  Biases:  ");
    for (v = 0; v<V; v++) printf(" %+8.3lf",gb[v]);
    printf("\n\n");
  }
  for (f = 0; f<F; f++)
  { if (f==0) printf("  Weights: ");
    else      printf("           ");
    for (v = 0; v<V; v++) printf(" %+8.3lf",gw[v*F+f]);
    printf("\n");
  }
  printf("\n");
  printf("  Noise var:");
  for (v = 0; v<V; v++) printf("%8.3lf ",gn[v]);
  printf("\n\n");

  printf("Recognition model:\n\n");
  if (!norm)
  { printf("  Biases:  ");
    for (f = 0; f<F; f++) printf(" %+8.3lf",rb[f]);
    printf("\n\n");
  }
  for (v = 0; v<V; v++)
  { if (v==0) printf("  Weights: ");
    else      printf("           ");
    for (f = 0; f<F; f++) printf(" %+8.3lf",rw[v*F+f]);
    printf("\n");
  }
  printf("\n");
  if (f>1 && corr)
  { for (f2 = 0; f2<F; f2++)
    { if (f2==0) printf("  Corr wts:");
      else       printf("           ");
      for (f = 0; f<F; f++)
      { if (f>f2) printf(" %+8.3lf",rc[f2*F+f]);
        else      printf("         ");
      }
      printf("\n");
    }
    printf("\n");
  }
  printf("  Noise var:");
  for (f = 0; f<F; f++) printf("%8.3lf ",rn[f]);
  printf("\n\n");
}


/* SAVE MODEL IN FILE AS AN S-PLUS EXPRESSION.  The model is an S-Plus
   "factanal" object.  It can be read using the S-Plus "source" function. */

static void save_model
( char *file
)
{
  FILE *fs;
  int v, f;

  fs = fopen(file,"w");

  if (fs==NULL)
  { fprintf(stderr,"Can't create model file: %s\n",model_file);
    exit(1);
  }

  fprintf(fs,"(function(){r<-list(\n");

  fprintf(fs,"  center = c(\n");
  for (v = 0; v<V; v++) 
  { fprintf(fs,"    %.6lf%s\n", mean[v], v==V-1 ? ")," : ",");
  }

  fprintf(fs,"  scale = c(\n");
  for (v = 0; v<V; v++) 
  { fprintf(fs,"    %.6lf%s\n", scale[v], v==V-1 ? ")," : ",");
  }

  fprintf(fs,"  uniquenesses = c(\n");
  for (v = 0; v<V; v++) 
  { fprintf(fs,"    %.6lf%s\n", gn[v], v==V-1 ? ")," : ",");
  }

  fprintf(fs,"  factors = %d,\n",F);

  fprintf(fs,"  loadings = matrix(c(\n");
  for (f = 0; f<F; f++)
  { fprintf(fs,"   ");
    for (v = 0; v<V; v++) 
    { fprintf(fs," %.6lf%s", gw[v*F+f], v==V-1 && f==F-1? ")," : ",");
    }
    fprintf(fs,"\n");
  }

  fprintf(fs,"  %d,%d)\n",V,F);

  fprintf(fs,"); class(r)<-\"factanal\"; r})()\n");

  fclose(fs);
}


/* DISPLAY USAGE MESSAGE AND EXIT. */

static void usage(void)
{
  fprintf(stderr,
   "Usage: ws-fa { options } vars factors iter-counts weight-step var-decay [ seed ]\n");
  fprintf(stderr,
   "             data-source [ model-file [ var-trace [ weight-trace ] ] ]\n");
  fprintf(stderr,
   "Options: -corr -rnoise -gnoise -norm -syst -echo\n");

  exit(1);
}
